from typing import Optional, Union

from pathlib import Path

import richfile as rf

class RichFile_data(rf.RichFile):
    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        check: Optional[bool] = True,
        safe_save: Optional[bool] = True,
    ):
        super().__init__(path=path, check=check, safe_save=safe_save)

        type_dicts = []
        
        ## NUMPY ARRAY
        try:
            import numpy as np

            def save_npy_array(
                obj: np.ndarray,
                path: Union[str, Path],
                **kwargs,
            ) -> None:
                """
                Saves a NumPy array to the given path.
                """
                np.save(path, obj, **kwargs)

            def load_npy_array(
                path: Union[str, Path],
                **kwargs,
            ) -> np.ndarray:
                """
                Loads an array from the given path.
                """    
                return np.load(path, **kwargs)
            
            type_dicts.append(
                {
                    "type_name":          "numpy_array",
                    "function_load":      load_npy_array,
                    "function_save":      save_npy_array,
                    "object_class":       np.ndarray,
                    "suffix":             "npy",
                    "library":            "numpy",
                    "versions_supported": [],
                },
                {
                    "type_name":          "numpy_scalar",
                    "function_load":      load_npy_array,
                    "function_save":      save_npy_array,
                    "object_class":       np.number,
                    "suffix":             "npy",
                    "library":            "numpy",
                    "versions_supported": [],
                },
            )
        except ImportError:
            pass
        

        ## SCIPY SPARSE MATRIX
        try:
            import scipy.sparse

            def save_sparse_array(
                obj: scipy.sparse.spmatrix,
                path: Union[str, Path],
                **kwargs,
            ) -> None:
                """
                Saves a SciPy sparse matrix to the given path.
                """
                scipy.sparse.save_npz(path, obj, **kwargs)

            def load_sparse_array(
                path: Union[str, Path],
                **kwargs,
            ) -> scipy.sparse.csr_matrix:
                """
                Loads a sparse array from the given path.
                """        
                return scipy.sparse.load_npz(path, **kwargs)
            
            type_dicts.append(
                {
                    "type_name":          "scipy_sparse_array",
                    "function_load":      load_sparse_array,
                    "function_save":      save_sparse_array,
                    "object_class":       scipy.sparse.spmatrix,
                    "suffix":             "npz",
                    "library":            "scipy",
                    "versions_supported": [],
                },
            )
        except ImportError:
            pass
        
        ## TORCH TENSOR
        try:
            import torch

            def save_torch_tensor(
                obj: torch.Tensor,
                path: Union[str, Path],
                **kwargs,
            ) -> None:
                """
                Saves a PyTorch tensor to the given path as a NumPy array.
                """
                np.save(path, obj.detach().cpu().numpy(), **kwargs)

            def load_torch_tensor(
                path: Union[str, Path],
                **kwargs,
            ) -> torch.Tensor:
                """
                Loads a PyTorch tensor from the given path.
                """
                return torch.from_numpy(np.load(path, **kwargs))
            
            type_dicts.append(
                {
                    "type_name":          "torch_tensor",
                    "function_load":      load_torch_tensor,
                    "function_save":      save_torch_tensor,
                    "object_class":       torch.Tensor,
                    "suffix":             "npy",
                    "library":            "torch",
                    "versions_supported": [],
                },
            )
        except ImportError:
            pass

        ## PANDAS DATAFRAME
        try:
            import pandas as pd
            
            def save_pandas_dataframe(
                obj: pd.DataFrame,
                path: Union[str, Path],
                **kwargs,
            ) -> None:
                """
                Saves a Pandas DataFrame to the given path.
                """
                ## Save as a CSV file
                obj.to_csv(path, index=True, **kwargs)

            def load_pandas_dataframe(
                path: Union[str, Path],
                **kwargs,
            ) -> pd.DataFrame:
                """
                Loads a Pandas DataFrame from the given path.
                """
                ## Load as a CSV file
                return pd.read_csv(path, index_col=0, **kwargs)
            
            type_dicts.append(
                {
                    "type_name":          "pandas_dataframe",
                    "function_load":      load_pandas_dataframe,
                    "function_save":      save_pandas_dataframe,
                    "object_class":       pd.DataFrame,
                    "suffix":             "csv",
                    "library":            "pandas",
                    "versions_supported": [],
                },
            )
        except ImportError:
            pass

        ## JSON DICT
        import collections
        import json

        def save_json_dict(
            obj: collections.UserDict,
            path: Union[str, Path],
            **kwargs,
        ) -> None:
            """
            Saves a dictionary to the given path.
            """
            with open(path, 'w') as f:
                json.dump(dict(obj), f, **kwargs)

        def load_json_dict(
            path: Union[str, Path],
            **kwargs,
        ) -> collections.UserDict:
            """
            Loads a dictionary from the given path.
            """
            with open(path, 'r') as f:
                return JSON_Dict(json.load(f, **kwargs))

        ## JSON LIST   
        def save_json_list(
            obj: collections.UserList,
            path: Union[str, Path],
            **kwargs,
        ) -> None:
            """
            Saves a list to the given path.
            """
            with open(path, 'w') as f:
                json.dump(list(obj), f, **kwargs)

        def load_json_list(
            path: Union[str, Path],
            **kwargs,
        ) -> collections.UserList:
            """
            Loads a list from the given path.
            """
            with open(path, 'r') as f:
                return JSON_List(json.load(f, **kwargs))
            
        type_dicts.append(
            {
                "type_name":          "json_dict",
                "function_load":      load_json_dict,
                "function_save":      save_json_dict,
                "object_class":       JSON_Dict,
                "suffix":             "json",
                "library":            "python",
                "versions_supported": [],
            },
            {
                "type_name":          "json_list",
                "function_load":      load_json_list,
                "function_save":      save_json_list,
                "object_class":       JSON_List,
                "suffix":             "json",
                "library":            "python",
                "versions_supported": [],
            },
        )

        [self.register_type_from_dict(d) for d in type_dicts]
        

######################################
######## CUSTOM DATA CLASSES #########
######################################

class JSON_Dict(dict):
    def __init__(self, *args, **kwargs):
        super(JSON_Dict, self).__init__(*args, **kwargs)
class JSON_List(list):
    def __init__(self, *args, **kwargs):
        super(JSON_List, self).__init__(*args, **kwargs)



f"""
RichFile

A system for saving and loading nested python objects containing array data
without serialization.

This module provides a useful way to deal with large and complex nested data
structures. It is designed to be: - insensitive to version changes in libraries
(unlike serialization / pickling) - accessible via browsing the folder structure
- fast and memory-mappable - customizable for different data types

The system is based on the following principles: 
- Each leaf object is saved as a separate file 
- The folder structure mirrors the nested object structure:
    - Lists, tuples, sets, frozensets are saved as folders with elements saved as files
      or folders with integer names
    - Dicts are saved as folders with items saved as folders with integer names.
      Dict items are saved as folders containing 2 elements.
- There is a single metadata file for each folder describing the properties of
  each element in the folder
    - The metadata file is a JSON file named ".metadata.richfile" and contains
      the following items:
        - "elements": a dictionary with keys that are the names of the files /
          folders in the directory and values that are dictionaries with the
          following items:
            - "type": A string describing type of the element. The string used
              should be a valid richfile type, as it is determines how the
              element is loaded. Examples: "npy_array", "scipy_sparse_array",
              "list", "object", "float", etc.
            - "library": A string describing the library used to save the
              element. Examples: "numpy", "scipy", "python", "json" (for native
              python types), etc.
           - "version": A string describing the version of the library used to
              save the element. This is used to determine how the element is
              loaded. Examples: "1.0.0", "0.1.0", etc.
            - "index": An integer that is used to determine the order of the
              elements when loading them. Example: 0, 1, 2, etc.
        - "type": A string describing the type of the folder. The string used
          should be a valid richfile type, as it determines how the folder is
          loaded. Examples: "list", "dict", "tuple", etc. (Only container-like
          types)
        - "library": A string describing the library used to save the folder.
          Examples: "python"
        - "version": A string describing the version of the library used to for
          the container. This is used to determine how the folder is loaded.
          Examples: "3.12", "3.13", etc.
        - "version_richfile": A string describing the version of the richfile
          format used to save the metadata file. Examples: "1.0.0", "0.1.0",
          etc.
- Loading proceeds as follows:
    - enter outer folder
    - load metadata file
    - check that files / folders in the directory match the metadata
    - if folder represents a list, tuple, set, frozenset:
        - elements are expected to be named as integers with an appropriate
          suffix: 0.list, 1.npy, 2.dict, 3.npz, 4.json, etc.
        - load each element in the order specified by the metadata index
        - if an element is container-like, enter its folder, load, and package
          it.
    - if folder represents a dict:
        - each item will be saved as a folder containing a single dict item
        - each dict item folder will contain 2 elements: key (0) and value (1)
    - load elements:
        - richfile types (eg. "array", "sparse_array", etc.) are saved and
          loaded using numpy, scipy, etc. as appropriate.
        - an appropriate suffix will be added to the file or folder name.
        - native python types (eg. "float", "int", "str", etc.) are saved as
          JSON files and loaded using the json library.          

RH 2024
"""

from typing import Any, Dict, List, Optional, Tuple, Union, Callable

from pathlib import Path
import json
import packaging
import packaging.version
import packaging.specifiers
import warnings
import functools
import importlib.metadata
import platform
import copy
import inspect
import shutil
import filelock
import os
from contextlib import ExitStack
import unicodedata
import sys

from . import functions
from . import __version__ as VERSION_RICHFILE
from . import VERSIONS_RICHFILE_SUPPORTED, PYTHON_VERSIONS_SUPPORTED, FILENAME_METADATA, FILENAME_TYPELOOKUP, JSON_INDENT, WINDOWS_RESERVED_NAMES


REQUIREMENTS = {
    ## Required top-level keys for the metadata file
    "keys_metadata": [
        "elements",
        "type",
        "library",
        "version",
        "version_richfile",
    ],
    ## Required keys for each element in the metadata file
    "keys_element": [
        "type",
        "library",
        "version",
        "index",
    ],
}


####################################################################################################
#################################### LOADING FUNCTIONS #############################################
####################################################################################################

def load_element(
    path: Union[str, Path],
    metadata: Dict,
    type_lookup: Dict,
    check: bool = True,
) -> Any:
    """
    Loads an element from the given path.
    """
    ## Elements can be either folders or files. The metadata determines how to load them.
    ### Go through a switch-case based on the "type" and possibly "library" and "version" of the element.
    if check:
        if not Path(path).exists():
            raise FileNotFoundError(f"Path {path} not found.")
        if not metadata["type"] in type_lookup:
            raise ValueError(f"Type {metadata['type']} not supported.")

    _prepare_element_loading(
        path=path,
        metadata=metadata,
        type_lookup=type_lookup,
        check=check,
    )

    ## Get function_load
    function_load = type_lookup[metadata["type"]]["function_load"]
    ## Check parameters accepted by function: path, type_lookup, metadata, check
    sig = inspect.signature(function_load)
    args_available = {
        "path": path,
        "metadata": metadata,
        "type_lookup": type_lookup,
        "check": check,
    }
    argnames = list(set(sig.parameters.keys()) & set(args_available.keys()))
    kwargs_function = {argname: args_available[argname] for argname in argnames}

    return function_load(
        **kwargs_function,
    )


def load_folder(
    path: Union[str, Path],
    type_lookup: Dict,
    check: bool = True,
    **kwargs,
) -> Any:
    """
    Loads the folder from the given path. Used for types: list, tuple, set, frozenset, dict.
    """
    metadata = load_folder_metadata(path_dir=path, name_metadata=FILENAME_METADATA, check=check)

    if check:
        names_path_elements = [p.name for p in Path(path).iterdir() if (p.name != FILENAME_METADATA) and (p.name != FILENAME_TYPELOOKUP)]
        # Check that index values are unique
        indices = [value["index"] for value in metadata["elements"].values()]
        if len(indices) != len(set(indices)):
            raise ValueError("Indices in metadata are not unique.")
        
        elif metadata["library"] == "python":
            if not is_version_compatible(version=metadata["version"], rules=PYTHON_VERSIONS_SUPPORTED):
                raise ValueError(f"Python version '{metadata['version']}' not supported for library 'python' and type '{metadata['type']}'. Version rules: {PYTHON_VERSIONS_SUPPORTED}")
        else:
            raise ValueError("Only 'python' library supported for container types.")
        
    ## Sort the element names by their index
    names_meta_sorted = _sort_element_names_by_index(metadata=metadata, check=check)

    if check:
        # Check that all elements in metadata are present in the folder and vice versa
        if not all(name in names_path_elements for name in names_meta_sorted):
            missing_elements = set(names_meta_sorted) - set(names_path_elements)
            raise FileNotFoundError(f"Elements in metadata not found in folder: {missing_elements}, names_meta: {names_meta_sorted}, names_path: {names_path_elements}, path: {path}")
        if not all(name in names_meta_sorted for name in names_path_elements):
            extra_elements = set(names_path_elements) - set(names_meta_sorted)
            raise ValueError(f"Extra elements in folder not found in metadata: {extra_elements}")        

    ## Load each element in the order specified by the metadata index
    elements = [load_element(
        path=str(Path(path) / name),
        metadata=metadata["elements"][name],
        type_lookup=type_lookup,
        check=check,
    ) for name in names_meta_sorted]
    
    if metadata["type"] == "list":
        pass
    elif metadata["type"] == "tuple":
        elements = tuple(elements)
    elif metadata["type"] == "set":
        elements = set(elements)
    elif metadata["type"] == "frozenset":
        elements = frozenset(elements)
    elif metadata["type"] == "dict_item":
        ## Make sure that there are exactly 2 elements
        if check:
            if len(elements) != 2:
                raise ValueError(f"DictItem must contain exactly 2 elements. Found {len(elements)}.")
        elements = DictItem(key=elements[0], value=elements[1])
    else:  ## elif metadata["type"] == "dict"
        ## Make sure that all elements are DictItem types
        if check:
            if not all(isinstance(element, DictItem) for element in elements):
                raise TypeError("All elements in a dict must be of type DictItem.")
        try:
            elements = {element.key: element.value for element in elements}  ## Unpack the DictItems
        except AttributeError as e:
            raise ValueError(f"Error unpacking dict items: {e}")
    
    return elements


def _sort_element_names_by_index(metadata: Dict, check: bool = True) -> List[str]:
    """
    Sorts the element names in the metadata by their index.
    """
    indices_and_names_meta = {value["index"]: name for name, value in metadata["elements"].items()}
    indices_meta_sorted = sorted(indices_and_names_meta.keys())
    names_meta_sorted = [indices_and_names_meta[index] for index in indices_meta_sorted]
    if check:
        if indices_meta_sorted != list(range(len(indices_meta_sorted))):
            raise ValueError("Indices in metadata are not consecutive integers starting from zero.")
    return names_meta_sorted


def load_folder_metadata(
    path_dir: Union[str, Path],
    name_metadata: str = FILENAME_METADATA,
    check: bool = True,
) -> Dict:
    """
    Loads the metadata file from the given path.
    """
    if check:
        if not Path(path_dir).is_dir():
            raise NotADirectoryError(f"Path {path_dir} is not a directory.")
        if not (Path(path_dir) / name_metadata).is_file():
            raise FileNotFoundError(f"Metadata file {name_metadata} not found in directory {path_dir}.")

    with open(str(Path(path_dir) / name_metadata), "r") as f:
        metadata = json.load(f)

    if check:
        # Check for required keys
        missing_keys = set(REQUIREMENTS["keys_metadata"]) - set(metadata.keys())
        if missing_keys:
            raise KeyError(f"Metadata is missing required keys: {missing_keys}")
        # Check version
        if not is_version_compatible(version=metadata["version_richfile"], rules=VERSIONS_RICHFILE_SUPPORTED):
            raise ValueError(f"RichFile version {metadata['version_richfile']} not supported.")
        for element_name, meta_element in metadata["elements"].items():
            missing_keys = set(REQUIREMENTS["keys_element"]) - set(meta_element.keys())
            if missing_keys:
                raise KeyError(f"Element '{element_name}' is missing keys: {missing_keys}")

    return metadata


def _prepare_element_loading(
    path: Union[str, Path],
    metadata: Dict,
    type_lookup: Dict,
    check: bool,
) -> None:
    """
    Performs checks and preparations for loading a file.
    """
    if check:
        if type_lookup[metadata["type"]]["library"] == []:
            warnings.warn(f"Field 'versions_supported' is empty in type_lookup for type {metadata['type']}.")
        ## If the library is python, check the variable directory
        elif metadata["library"] == "python":
            if not is_version_compatible(version=metadata["version"], rules=PYTHON_VERSIONS_SUPPORTED):
                raise ValueError(f"Python version '{metadata['version']}' not supported for library 'python' and type '{metadata['type']}'. Version rules: {PYTHON_VERSIONS_SUPPORTED}")
        elif not is_version_compatible(version=metadata["version"], rules=type_lookup[metadata["type"]]["versions_supported"]):
            raise ValueError(f"Version '{metadata['version']}' not supported for library '{metadata['library']}' and type '{metadata['type']}'. Version rules: {type_lookup[metadata['type']]['versions_supported']}")
        ## Check that path exists as either a file or a directory
        if not Path(path).exists():
            raise FileNotFoundError(f"Path {path} not found.")            

def load_json(
    path: Union[str, Path],
    **kwargs,
) -> Any:
    """
    Loads a scalar from the given path. Used as a super for str, bool, and None.
    """
    with open(path, "r") as f:
        return json.load(f, **kwargs)
    
def load_float(path: Union[str, Path], **kwargs) -> float:
    return float(load_json(path, **kwargs))
def load_int(path: Union[str, Path], **kwargs) -> int:
    return int(load_json(path, **kwargs))
def load_str(path: Union[str, Path], **kwargs) -> str:
    return str(load_json(path, **kwargs))
def load_bool(path: Union[str, Path], **kwargs) -> bool:
    return bool(load_json(path, **kwargs))
def load_None(path: Union[str, Path], **kwargs) -> None:
    out = load_json(path, **kwargs)
    if out is not None:
        raise ValueError("Loaded object is not None.")
    return out
    

####################################################################################################
#################################### SAVING FUNCTIONS ##############################################
####################################################################################################


def save_object(
    obj: Any,
    path: Union[str, Path],
    type_lookup: Dict,
    check: bool = True,
    overwrite: bool = False,
    name_dict_items: bool = True,
) -> None:
    """
    Saves an object to the given directory in the RichFile format.
    """
    # Determine the type of the object and save accordingly
    props = type_lookup[type(obj)]
    type_object = props["type_name"]

    ## Make directory if it doesn't exist
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    
    if check:
        ## Check illegal Windows folder names
        stem = Path(path).stem
        if stem.upper() in WINDOWS_RESERVED_NAMES:
            raise ValueError((f"ERROR: Cannot save folder named with stem: '{stem}': "
                "Windows reserves names like COM1-9, LPT1-9, CON, PRN, AUX, NUL, etc.  "
                "Either rename your file/dict key or set name_dict_items=False or set check=False."))
        
        ## Check for overwriting
        if Path(path).exists() and not overwrite:
            raise FileExistsError(f"Path already exists: {path}")

        library_version = _get_library_version(props["library"])
        if props["library"] in type_lookup:
            if not is_version_compatible(version=library_version, rules=type_lookup[props["library"]]["versions_supported"]):
                raise ValueError(f"Library {props['library']} version {library_version} not supported.")        

    ## Get function_save
    function_save = props["function_save"]
    ## Check parameters accepted by function: obj, path, type_lookup, type_name, check, overwrite, name_dict_items
    sig = inspect.signature(function_save)
    args_available = {
        "obj": obj,
        "path": path,
        "type_lookup": type_lookup,
        "type_name": type_object,
        "check": check,
        "overwrite": overwrite,
        "name_dict_items": name_dict_items,
    }
    argnames = list(set(sig.parameters.keys()) & set(args_available.keys()))
    kwargs_function = {argname: args_available[argname] for argname in argnames}

    function_save(
        **kwargs_function,
    )

def save_container(
    obj: Union[list, tuple, set, dict, frozenset, 'DictItem'],
    path: Union[str, Path],
    type_name: str,
    type_lookup: Dict,
    check: bool = True,
    overwrite: bool = False,
    name_dict_items: bool = True,
) -> None:
    """
    Saves a list, tuple, set, frozenset, or dict_item to the given directory.
    """
    if isinstance(obj, dict):
        _check_case_only_sibling_keys(obj=obj) if check and (name_dict_items) else None 
        obj = [DictItem(key=key, value=value) for key, value in obj.items()]

    metadata_elements = {}
    for idx, element in enumerate(obj):
        try:
            props = type_lookup[type(element)]
            type_element = props["type_name"]            
        except TypeError as e:
            raise TypeError(f"Failed to get properties for element. \n Directory: {path}. \n Index: {idx}. \n element: {element}. \n type_name: {type_name}. \n Error: {e}")
        name_element = f"{idx}.{props['suffix']}"  ## Make name the index and add suffix
        if name_dict_items:
            if type_element == "dict_item":
                if isinstance(element.key, str):
                    name_element = f"{element.key}.{props['suffix']}"
            elif type_name == "dict_item":
                name_element = f"{['key', 'value'][idx]}.{props['suffix']}"

        _check_filename_safety(name=name_element, warn=True, raise_error=False)
        path_element = str(Path(path) / name_element)  ## Make path
        save_object(
            obj=element,
            path=path_element,
            check=check,
            overwrite=overwrite,
            type_lookup=type_lookup,
        )
        metadata_elements[name_element] = {
            "type": type_element,
            "library": props["library"],
            "version": _get_library_version(library=props["library"]),
            "index": idx,
        }

    metadata_container = {
        "elements": metadata_elements,
        "type": type_name,
        "library": "python",
        "version": _get_python_version(),
        "version_richfile": VERSION_RICHFILE,
    }
    save_metadata(
        metadata=metadata_container,
        path_dir=path,
        check=check,
        name_metadata=FILENAME_METADATA,
        overwrite=overwrite,
    )


def _check_filename_safety(name: str, warn: bool = True, raise_error: bool = False) -> None:
    """
    Checks if a filename is safe to use.
    """
    issues = invalid_chars_filename(name)
    bad_chars = [c for c, valid in issues.items() if not valid]
    n_issues = len(bad_chars)

    if n_issues > 0:
        if warn:
            warnings.warn(f"Filename contains invalid character: {bad_chars}. Name: {name}")
        if raise_error:
            raise ValueError(f"Filename contains invalid character: {bad_chars}. Name: {name}")

def _check_case_only_sibling_keys(obj: Dict) -> None:
    """
    Checks if there are any case-only sibling keys in the dictionary.
    """
    ## Check that there are no case-only sibling keys
    str_keys = [k for k in obj.keys() if isinstance(k, str)]
    str_keys_lower = [key.lower() for key in str_keys]
    case_only_siblings = [key for key in str_keys if str_keys_lower.count(key.lower()) > 1]
    if len(case_only_siblings) > 0:
        raise ValueError((f"ERROR: Found case-only sibling keys in dict: '{obj}'."
                        f"Tried to save dict with keys '{str_keys}' but found case-only sibling(s) named {[str(c) for c in case_only_siblings]}. "
                        "These are illegal on MacOS and Windows because their filesystems are case-insensitive."
                        "Either rename your file/dict key or set name_dict_items=False or set check=False."))


def save_metadata(
    metadata: Dict,
    path_dir: Union[str, Path],
    check: bool = True,
    name_metadata: str = FILENAME_METADATA,
    overwrite: bool = False,
) -> None:
    """
    Saves the metadata dictionary to a file in the given directory.
    """
    Path(path_dir).mkdir(parents=True, exist_ok=True)
    if check:
        ## There should not be a metadata file already present if overwrite is False
        if not overwrite and (Path(path_dir) / name_metadata).exists():
            raise FileExistsError(f"Metadata file {name_metadata} already exists in directory {path_dir}.")

    with open(str(Path(path_dir) / name_metadata), "w") as f:
        json.dump(metadata, f, indent=JSON_INDENT)


class DictItem:
    def __init__(self, key, value):
        self.key = key
        self.value = value
    def __iter__(self):
        return iter([self.key, self.value])

def save_json(
    obj: Any,
    path: Union[str, Path],
    **kwargs,
) -> None:
    """
    Saves a JSON-serializable object to the given path.
    """
    with open(path, "w") as f:
        json.dump(obj, f)


def _get_library_version(library: str) -> str:
    """
    Returns the version of the library as a string.
    """
    ## Import the library str's version using importlib.metadata
    ### If it's a native python library, use a custom function to get the version
    if library in ["python", "builtins"]:
        return _get_python_version()
    else:
        try:
            return importlib.metadata.version(library)
        except importlib.metadata.PackageNotFoundError:
            try:
                ## Try to import the library and get the version from the __version__ attribute
                lib = importlib.import_module(library)
            except AttributeError as e:
                raise ValueError(f"Library: '{library}' not found. Error: {e}")
            try:
                return lib.__version__
            except AttributeError as e:
                warnings.warn(f"Library {library} does not have a __version__ attribute. Error: {e}")
                return None
        except ImportError:
            raise ValueError(f"Library {library} not found. Error: {e}")
    
def _get_python_version() -> str:
    """
    Returns the Python version as a string.
    """
    return platform.python_version()

def is_version_compatible(version: str, rules: List[str]) -> bool:
    """
    Checks if a given version string satisfies all specified version rules.
    RH 2024

    Args:
        version (str): 
            The version string to check (e.g., "2.4.9").
        rules (List[str]): 
            A list of version rules (e.g., ["<3", ">=1.2",]).

    Returns:
        bool: True if the version satisfies all rules, False otherwise.
    """
    try:
        version_obj = packaging.version.Version(version)
        specifier = packaging.specifiers.SpecifierSet(",".join(rules))
        return version_obj in specifier
    except packaging.version.InvalidVersion:
        raise ValueError(f"Invalid version string: {version}")


def is_container_type(obj: Any, type_lookup: Dict) -> bool:
    """
    Checks if an object is a container type by seeing if the save function is 'save_container'.
    """
    props = type_lookup[type(obj)]
    function_save = props["function_save"]
    if function_save.__code__.co_code == save_container.__code__.co_code:
        return True
    else:
        ## Try duck-typing to see if function_save ends up calling save_container
        try:
            ## Pass in dummy arguments and obj: empty list
            import tempfile
            with tempfile.TemporaryDirectory() as path:
                function_save(
                    obj=[],
                    path=path,
                    type_name='list',
                    type_lookup=type_lookup,
                    check=True,
                    overwrite=True,
                    name_dict_items=False,
                )
                ## If a file was created, return error
                if Path(path).is_file():
                    return FileNotFoundError(f"Folder created as file.")
                ## Check to see if a folder was created
                elif not Path(path).is_dir():
                    return FileNotFoundError(f"Folder not created.")
                ## If so, check to see if the metadata file was created
                elif not (Path(path) / FILENAME_METADATA).is_file():
                    return FileNotFoundError(f"Metadata file not created.") 
                ## If it worked
                else:               
                    return True
            ## Folder should be deleted after the with statement
        except Exception as e:
            return e
        

def invalid_chars_filename(name: str) -> str:
    """
    Checks if a filename contains invalid characters.
    """
    def single(c):
        code_point = ord(c)
        
        # Control characters (ASCII codes 0-31, 127)
        if code_point <= 31 or code_point == 127:
            return False
        
        # Invalid characters on Windows
        if c in '<>:"/\\|?*':
            return False
        
        # Path separator on Unix/Linux and macOS
        if c == '/':
            return False
        
        # Surrogate halves (U+D800 to U+DFFF)
        if 0xD800 <= code_point <= 0xDFFF:
            return False
        
        # Noncharacters (U+FDD0 to U+FDEF, and code points ending with FFFE or FFFF)
        if 0xFDD0 <= code_point <= 0xFDEF or (code_point & 0xFFFF) in (0xFFFE, 0xFFFF):
            return False
        
        # Code points beyond U+FFFF (outside BMP)
        if code_point > 0xFFFF:
            return False
        
        # Unassigned code points
        if unicodedata.category(c) == 'Cn':
            return False
        
        return True  

    return {c: single(c) for c in name if not single(c)}


####################################################################################################
#################################### HIGH-LEVEL CLASS ##############################################
####################################################################################################

class RichFile:
    f"""
    High-level class for handling reading and writing objects in the RichFile format.
    Allows customization of loading and saving functions, and setting additional parameters.
    RH 2024

    Args:
        path (Optional[Union[str, Path]): 
            The path to save the object to. If None, uses the path specified
            in the RichFile object.
        check (Optional[bool]): 
            Whether to perform checks on the object and path. If None, uses
            the check specified in the RichFile object.
        safe_save (Optional[bool]): 
            Whether to use a safe save method. If None, uses the safe_save
            specified in the RichFile object.
        overwrite (Optional[bool]): 
            Whether to overwrite the file if it already exists. If None,
            uses the overwrite specified in the RichFile object.
        name_dict_items (Optional[bool]):
            Whether to name dict items as their keys. If None, uses the
            name_dict_items specified in the RichFile object.
        save_type_lookup (Optional[bool]):
            Whether to save the type lookup table as a file named
            {FILENAME_TYPELOOKUP} in the outermost richfile directory if it
            is a directory. If None, uses the save_type_lookup specified in
            the RichFile object.
    """
    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        check: Optional[bool] = True,
        safe_save: Optional[bool] = True,
        overwrite: Optional[bool] = False,
        name_dict_items: Optional[bool] = True,
        save_type_lookup: Optional[bool] = True,
    ):
        self.path             = path
        self.check            = check
        self.safe_save        = safe_save
        self.overwrite        = overwrite
        self.name_dict_items  = name_dict_items
        self.save_type_lookup = save_type_lookup

        self.type_lookup = copy.deepcopy(functions.TypeLookup())
        self.params_load = {}
        self.params_save = {}
    
    def save(
        self, 
        obj: Any, 
        path: Optional[Union[str, Path]] = None, 
        check: Optional[bool] = None, 
        safe_save: Optional[bool] = None,
        overwrite: Optional[bool] = None,
        name_dict_items: Optional[bool] = None,
        save_type_lookup: Optional[bool] = None,
    ) -> None:
        f"""
        Saves an object to the given path.

        Args:
            obj (Any): 
                The object to save.
            path (Optional[Union[str, Path]]): 
                The path to save the object to. If None, uses the path specified
                in the RichFile object.
            check (Optional[bool]): 
                Whether to perform checks on the object and path. If None, uses
                the check specified in the RichFile object.
            safe_save (Optional[bool]): 
                Whether to use a safe save method. If None, uses the safe_save
                specified in the RichFile object.
            overwrite (Optional[bool]): 
                Whether to overwrite the file if it already exists. If None,
                uses the overwrite specified in the RichFile object.
            name_dict_items (Optional[bool]):
                Whether to name dict items as their keys. If None, uses the
                name_dict_items specified in the RichFile object.
            save_type_lookup (Optional[bool]):
                Whether to save the type lookup table as a file named
                {FILENAME_TYPELOOKUP} in the outermost richfile directory if it
                is a directory. If None, uses the save_type_lookup specified in
                the RichFile object.
        """
        path             = self.path             if path             is None else path
        check            = self.check            if check            is None else check
        safe_save        = self.safe_save        if safe_save        is None else safe_save
        overwrite        = self.overwrite        if overwrite        is None else overwrite
        name_dict_items  = self.name_dict_items  if name_dict_items  is None else name_dict_items
        save_type_lookup = self.save_type_lookup if save_type_lookup is None else save_type_lookup

        if (path is None) or (not isinstance(check, bool)):
            raise ValueError("`path` [str, Path] and `check` [bool] must be specified.")
        
        kwargs_safe_saver = {
            "overwrite": overwrite,
            "safe_save": safe_save,
            "delete_temp_on_error": False,
            "timeout_lock": 1,
            "force_acquire_lock": overwrite,
            "force_release_lock": True,
        }

        ## Make sure that the object is a container type
        is_continer = is_container_type(obj=obj, type_lookup=self.type_lookup)
        function_save = self.type_lookup[type(obj)]["function_save"]
        if is_continer != True:
            raise TypeError(f"Object must be a container type. \nObject failed `is_container_type` test. Found type: {type(obj)}, and `function_save`: {function_save}. \nConsider wrapping the object in a dict (or list, tuple, set, frozenset, or richfile.util.DictItem). \nError: {is_continer}")

        ## Create a lock file specific to the target path
        ### Append the lock suffix to the path (don't replace the suffix)
        path = str(path)
        fn_make_path_tmp  = lambda path: path + '.tmp'
        fn_make_path_lock = lambda path: path + '.lock'
        with SafeSaver(
            path_target=path,
            path_temp=fn_make_path_tmp(path),
            path_lock=fn_make_path_lock(path),
            **kwargs_safe_saver,
        ) as path_temp:
            save_object(
                obj=obj,
                path=path_temp,
                type_lookup=self.type_lookup,
                check=check,
                overwrite=overwrite,
                name_dict_items=name_dict_items,
            )
            
        if save_type_lookup:
            ## Make the type_lookup table a file in the outermost directory
            type_lookup = copy.deepcopy(self.type_lookup.properties)
            ### Convert the functions and classes into strings
            for prop in type_lookup:
                prop["function_load"] = inspect.getsource(prop["function_load"])
                prop["function_save"] = inspect.getsource(prop["function_save"])
            
                prop["object_class"] = str(prop["object_class"])

            ## If the richfile is a directory, save the type lookup table as a file in the outermost directory
            if Path(path).is_dir():
                ## Save as a .json file
                path_type_lookup = str(Path(path) / FILENAME_TYPELOOKUP)
                with SafeSaver(
                    path_target=path_type_lookup,
                    path_temp=fn_make_path_tmp(path_type_lookup),
                    path_lock=fn_make_path_lock(path_type_lookup),
                    **kwargs_safe_saver,
                ) as path_temp:
                    save_json(
                        obj=type_lookup,
                        path=path_temp,
                        indent=JSON_INDENT,
                    )
                
        return self
    
    def load(
        self, 
        path: Optional[Union[str, Path]] = None, 
        type_lookup: Optional[Dict] = None,
        check: Optional[bool] = None,
    ) -> Any:
        path = self.path if path is None else path
        type_lookup = self.type_lookup if type_lookup is None else type_lookup
        check = self.check if check is None else check
        if (path is None) or (not isinstance(check, bool)):
            raise ValueError("`path` [str, Path] and `check` [bool] must be specified.")
        
        ## Look for a metadata file in the directory
        ### If there isn't one, assume it is the outer directory and load as a folder
        if not (Path(path).parent / FILENAME_METADATA).exists():
            ## If the path is a directory, load it as a folder
            if Path(path).is_dir():
                ## Look for the type within the metadata file within the directory
                if not (Path(path) / FILENAME_METADATA).exists():
                    raise FileNotFoundError(f"Metadata file {FILENAME_METADATA} not found in directory {path}.")
                metadata_dir = load_folder_metadata(path_dir=path, check=check)
                metadata_obj = {
                    "type": metadata_dir["type"],
                    "library": metadata_dir["library"],
                    "version": metadata_dir["version"],
                    "index": 0,
                }
                return load_element(
                    path=path,
                    metadata=metadata_obj,
                    type_lookup=type_lookup,
                    check=check,
                )
            ## If the path is a file, then it is missing a metadata file. Try to load it using object properties
            elif Path(path).is_file():
                props = type_lookup[type(Path(path).suffix)]
                metadata_obj = {
                    "type": props["type_name"],
                    "library": props["library"],
                    "version": _get_library_version(library=props["library"]),
                    "index": 0,
                }
                return load_element(
                    path=path,
                    metadata=metadata_obj,
                    type_lookup=type_lookup,
                    check=check,
                )
            else:
                raise FileNotFoundError(f"Path {path} not found.")
        else:
            metadata_folder = load_folder_metadata(path_dir=str(Path(path).parent), check=check)
            metadata_element = metadata_folder["elements"][Path(path).name]
            return load_element(
                path=path,
                metadata=metadata_element,
                type_lookup=type_lookup,
                check=check,
            )

    def register_type(
        self, 
        type_name: str, 
        function_load: Callable, 
        function_save: Callable, 
        object_class: type, 
        suffix: str, 
        library: str,
        versions_supported: Optional[List[str]] = [],
    ):
        """
        Registers a new type with custom loading and saving functions.
        """
        ## Add the property to the type_lookup
        prop = {
            'type_name': type_name,
            'function_load': function_load,
            'function_save': function_save,
            'object_class': object_class,
            'suffix': suffix,
            'library': library,
            'versions_supported': versions_supported,
        }

        if self.check:
            functions._verify_validity_of_new_type(prop)
            #### check for duplicates
            if type_name in self.type_lookup:
                raise KeyError(f"Type {type_name} already registered.")

        self.type_lookup.add_property(prop)

    def register_type_from_dict(self, prop: Dict) -> None:
        """
        Registers a new type with custom loading and saving functions.
        """
        if self.check:
            if not isinstance(prop, dict):
                raise TypeError("`prop` must be a dictionary.")
        ## Add in versions_supported if not present
        prop.setdefault("versions_supported", [])

        ## Use inspect to ensure that all args in register_type are present. Kwargs okay to skip.
        sig = inspect.signature(self.register_type)
        args_available = set(prop.keys())
        args_missing = set(sig.parameters.keys()) - args_available
        args_extra = args_available - set(sig.parameters.keys())
        if len(args_missing) > 0:
            raise ValueError(f"Missing arguments: {args_missing}.")
        if len(args_extra) > 0:
            raise ValueError(f"Extra arguments: {args_extra}.")
        
        self.register_type(**prop)

    def remove_type(self, type_name: str) -> None:
        """
        Removes a type from the type lookup table.
        """
        if self.check:
            if type_name not in self.type_lookup:
                raise KeyError(f"Type {type_name} not found.")
        self.type_lookup.remove_property(type_name)

    def set_load_kwargs(
        self, 
        type_: Union[str, type], 
        **kwargs,
    ) -> None:
        """
        Sets additional parameters for the load function of a specific type.
        """
        type_name = self.type_lookup[type_]["type_name"]

        ## Partial the function with the parameters
        if type_name in self.type_lookup:
            self.type_lookup[type_name] = {
                "function_load": functools.partial(
                    self.type_lookup[type_name]['function_load'],
                    **kwargs,
                )
            }
        else:
            raise KeyError(f"Type {type_name} not found in type lookup.")
        
        self.params_load[type_name] = kwargs

    def set_save_kwargs(
        self,
        type_: Union[str, type], 
        **kwargs,
    ) -> None:
        """
        Sets additional parameters for the save function of a specific type.
        """
        type_name = self.type_lookup[type_]["type_name"]

        ## Partial the function with the parameters
        if type_name in self.type_lookup:
            self.type_lookup[type_name] = {
                "function_save": functools.partial(
                    self.type_lookup[type_name]['function_save'],
                    **kwargs,
                )
            }
        else:
            raise KeyError(f"Type {type_name} not found in type lookup.")
        
        self.params_save[type_name] = kwargs

    def get_metadata(self, path_dir: Union[str, Path]) -> Dict:
        """
        Retrieves the metadata from the specified directory.
        """
        return load_folder_metadata(path_dir=path_dir, check=self.check)

    def list_elements(self, path_dir: Union[str, Path]) -> List[str]:
        """
        Lists the elements in the specified directory.
        """
        metadata = load_folder_metadata(path_dir=path_dir, check=self.check)
        return list(metadata['elements'].keys())

    def get_type_info(self, type_name: str) -> Dict:
        """
        Retrieves the type information for a given type name.
        """
        return self.type_lookup.get(type_name, {})

    def view_directory_tree(self, path: Optional[Union[str, Path]] = None) -> None:
        """
        Prints a tree structure of the directory.
        Uses the metadata to determine the structure.
        """
        path = self.path if path is None else path
        if path is None:
            raise ValueError("`path` [str, Path] must be specified.")
        
        def _view_tree(path, level=0):
            metadata = self.get_metadata(path)
            for name, value in metadata['elements'].items():
                print("|   " * level + "├── " + f"{name} ({value['type']})")
                if value['type'] in ["list", "tuple", "set", "frozenset", "dict", "dict_item"]:
                    _view_tree(path=str(Path(path) / name), level=level+1)
            print("|   " * level)

        if Path(path).is_dir():
            print(f"Viewing tree structure of richfile at path: {path} ({self.get_metadata(path)['type']})")
            _view_tree(path)
        elif Path(path).is_file():
            metadata_folder = load_folder_metadata(path_dir=str(Path(path).parent), check=self.check)
            name_element = Path(path).name
            metadata_element = metadata_folder["elements"][name_element]
            print(f"Viewing element at path: {path} ({metadata_element['type']})")
        else:
            raise FileNotFoundError(f"Path {path} not found.")
    
    def view_tree(
        self, 
        path: Optional[Union[str, Path]] = None, 
        show_filenames: bool = False,
    ) -> None:
        """
        Prints a tree structure of the directory.
        If a dict item has a string key, it will be printed as a key-value pair.
        List, tuple, set, and frozenset items will be printed as a list of items.
        """
        sf = show_filenames
        path = self.path if path is None else path
        if path is None:
            raise ValueError("`path` [str, Path] must be specified.")
        
        def _view_tree(path, level=0):
            metadata = self.get_metadata(path)
            for name, value in metadata['elements'].items():
                if value['type'] == "dict_item":
                    metadata_item = self.get_metadata(str(Path(path) / name))
                    names_meta_sorted = _sort_element_names_by_index(metadata=metadata_item)
                    ## The other element should be the value
                    if len(names_meta_sorted) != 2:
                        raise ValueError("DictItem must contain exactly 2 elements: key='0.json' and value=another element.")
                    name_key, name_value = names_meta_sorted
                    metadata_key = metadata_item['elements'][name_key]
                    metadata_value = metadata_item['elements'][name_value]
                    key = load_element(
                        path=str(Path(path) / name / name_key),
                        metadata=metadata_key,
                        type_lookup=self.type_lookup,
                    )
                    print("|    " * level + "├── " + f"'{key}': {(name_value if sf else '')}  ({metadata_value['type']})")
                    if metadata_value['type'] in ["list", "tuple", "set", "dict", "dict_item"]:
                        _view_tree(path=str(Path(path) / name / name_value), level=level+1)
                elif value['type'] in ["list", "tuple", "set", "dict"]:
                    print("|    " * level + "├── " + f"{(name if sf else '')}  ({value['type']})")
                    _view_tree(path=str(Path(path) / name), level=level+1)
                else:
                    print("|    " * level + "├── " + f"{(name if sf else '')}  ({value['type']})")
            print("|    " * level)

        if Path(path).is_dir():
            print(f"Path: {path} ({self.get_metadata(path)['type']})")
            _view_tree(path)
        elif Path(path).is_file():
            metadata_folder = load_folder_metadata(path_dir=str(Path(path).parent), check=self.check)
            name_element = Path(path).name
            metadata_element = metadata_folder["elements"][name_element]
            print(f"Path: {path} ({metadata_element['type']})")
        else:
            raise FileNotFoundError(f"Path {path} not found.")

    def get_metadata_tree(
        self,
        path: Optional[Union[str, Path]] = None,
    ):
        """
        Return a hierarchical dictionary containing the metadata for the entire
        directory tree.
        """
        path = self.path if path is None else path
        if path is None:
            raise ValueError("`path` [str, Path] must be specified.")
        
        def _get_metadata_tree(path):
            metadata = self.get_metadata(path)
            out = {
                "metadata": metadata,
                "elements": {},
            }
            for name, value in metadata['elements'].items():
                if value['type'] in ["list", "tuple", "set", "frozenset", "dict", "dict_item"]:
                    out["elements"][name] = _get_metadata_tree(str(Path(path) / name))
                else:
                    out["elements"][name] = value
            return out

        if Path(path).is_dir():
            return _get_metadata_tree(path)
        elif Path(path).is_file():
            metadata_folder = load_folder_metadata(path_dir=str(Path(path).parent), check=self.check)
            name_element = Path(path).name
            metadata_element = metadata_folder["elements"][name_element]
            return metadata_element
        else:
            raise FileNotFoundError(f"Path {path} not found.")
        
    def __str__(self):
        return f"RichFile(path={self.path}, check={self.check}, params_load={self.params_load}, params_save={self.params_save})"

    def __repr__(self):
        ## If the path exists and has a metadata file, show the metadata
        if self.path is not None:
            if Path(self.path).exists() and (Path(self.path) / FILENAME_METADATA).exists():
                metadata = self.get_metadata(self.path)
                self.view_tree()

        return self.__str__()
    
    ## Item retrieval by key or index
    def __getitem__(self, key):
        
        ## Load dict items by key
        if isinstance(key, str):
            metadata = self.get_metadata(self.path)
            ## Confirm that path is a dict
            if metadata['type'] != "dict":
                raise ValueError("Path must be a dict to load by key.")
            ## Find filename matching a dict_item with a str as a key matching the input key
            names_meta_sorted = _sort_element_names_by_index(metadata=metadata)
            for name in names_meta_sorted:
                ## Check if the key is a string by loading the dict_item metadata
                metadata_item = self.get_metadata(str(Path(self.path) / name))
                if (not (metadata['elements'][name]['type'] == "dict_item")):
                    raise ValueError(f"Found element with type {metadata['elements'][name]['type']}. Expected 'dict_item'.")
                if not (metadata_item['type'] == "dict_item"):
                    raise ValueError(f"Found element with type {metadata_item['type']}. Expected 'dict_item'.")
                if not len(metadata_item['elements']) == 2:
                    raise ValueError("DictItem must contain exactly 2 elements: key and value.")
                names_meta_sorted_item = _sort_element_names_by_index(metadata=metadata_item)
                name_key, name_value = names_meta_sorted_item
                if metadata_item['elements'][name_key]['type'] == "str":
                    key_loaded = load_element(
                        path=str(Path(self.path) / name / name_key),
                        metadata=metadata_item['elements'][name_key],
                        type_lookup=self.type_lookup,
                    )
                    if key_loaded == key:
                        ## return a RichFile object at the path of the value
                        out = copy.deepcopy(self)
                        out.path = str(Path(self.path) / name / name_value)
                        return out
                        
        ## Load list or tuple items by index
        elif isinstance(key, int):
            metadata = self.get_metadata(self.path)
            ## Confirm that path is a list or tuple
            if metadata['type'] not in ["list", "tuple"]:
                raise ValueError("Path must be a list or tuple to load by index.")
            ## Find filename of the metadata index matching the input key
            for name, value in metadata['elements'].items():
                if value['index'] == key:
                    ## return a RichFile object at the path of the element
                    out = copy.deepcopy(self)
                    out.path = str(Path(self.path) / name)
                    return out
        else:
            raise ValueError("__getitem__ only supports str and int keys.")
        
        raise KeyError(f"Key {key} not found.")
    
    def keys(self):
        """
        Returns a list of keys in the directory.
        """
        try:
            metadata = self.get_metadata(self.path)
            names_elements_raw = list(metadata['elements'].keys())
            names_elements = ['.'.join(name.split('.')[:-1]) for name in names_elements_raw]
            return names_elements
        except FileNotFoundError:
            return []
        except Exception as e:
            warnings.warn(f"Path element failed to load metadata or doesn't have .keys() method. Error: {e}")
    

def delete_file_or_folder(path: Union[str, Path]) -> None:
    """
    Deletes files OR directories.

    Args:
        path (Union[str, Path]): 
            The path to the file or directory to delete.
    """
    path = Path(path)
    try:
        if path.is_file() or path.is_symlink():
            path.unlink()
        elif path.is_dir():
            shutil.rmtree(str(path))
        else:
            pass  ## path does not exist, do nothing
    except Exception as e:
        warnings.warn(f"Failed to delete path: {path}. Error: {e}")
    

class FileLock_AutoDeleting(filelock.FileLock):
    """
    A robust and safe wrapper around filelock.FileLock that ensures the lock
    file is deleted upon release or when exiting a with statement, preventing
    orphaned lock files.

    Args:
        lock_file (Union[str, Path]):
            The path to the lock file.
        timeout (Optional[float]):
            The maximum time to wait for the lock in seconds.
        force_acquire (bool):
            If True, the lock will be acquired even if the lock file exists.
        force (bool):
            NOT USED. This class always forces the lock to be released and
            deleted.
    """
    def __init__(
        self, 
        lock_file: Union[str, Path], 
        timeout: Optional[float] = None,
        force_acquire: bool = False,
        force: bool = False,
    ):
        if isinstance(force_acquire, bool):
            self.force_acquire = force_acquire
        else:
            raise ValueError("`force_acquire` must be a boolean.")

        if force_acquire and Path(lock_file).exists():
            Path(lock_file).unlink()

        self.path_lock = str(lock_file)
        super().__init__(lock_file=str(lock_file), timeout=timeout)

    def release(self, force: bool = False):
        """
        Release the lock and delete the lock file if it exists.
        """
        super().release(force=force)
        try:
            Path(self.path_lock).unlink()
        except FileNotFoundError:
            pass
        except Exception as e:
            warnings.warn(f"Failed to remove lock file: {e}")

    def __exit__(self, exc_type, exc_value, traceback):
        self.release(force=False)
        return False
    
    def __del__(self):
        self.release(force=False)

        
class AtomicSaver:
    """
    Context manager for safely saving files or directories using a temporary
    path. On successful exit, it replaces the target path with the temporary
    data. On error, it deletes the temporary path if desired.
    RH 2024

    Args:
        path_target (Union[str, Path]):
            The path to the target file or directory to save.
        path_temp (Optional[Union[str, Path]):
            The path to the temporary file or directory to save. If None, it
            will be the target path with '.tmp' appended.
        overwrite (bool):
            If True, the target path will be overwritten if it exists.
        delete_temp_on_error (bool):
            If True, the temporary path will be deleted if an error occurs.    
    """
    def __init__(
        self, 
        path_target: Union[str, Path], 
        path_temp: Optional[Union[str, Path]] = None,
        overwrite: bool = False,
        delete_temp_on_error: bool = False,
    ):
        self.path_target = str(path_target)
        self.path_temp = str(path_temp) if path_temp is not None else self.path_target + '.tmp'

        self.overwrite = overwrite
        self.delete_temp_on_error = delete_temp_on_error

    def __enter__(self):
        ## Handle overwrite protection
        if not self.overwrite:
            if Path(self.path_target).exists():
                raise FileExistsError(f"Path already exists: {self.path_target} and overwrite is False.")
        ## Delete the temp path if it exists
        if Path(self.path_temp).exists():
            delete_file_or_folder(self.path_temp)
            
        return self.path_temp

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            if exc_type is None:
            ## No exception, proceed to replace target with temp
                ## Replace the target path with the temp path atomically using os.replace
                ### Handle non-empty directories by deleting the target path first (loss of atomicity)
                if Path(self.path_target).is_dir():
                    if Path(self.path_target).exists():
                        delete_file_or_folder(self.path_target)
                os.replace(src=self.path_temp, dst=self.path_target)
                
            else:
                ## An exception occurred, delete the temp path
                if Path(self.path_temp).exists() and self.delete_temp_on_error:
                    delete_file_or_folder(self.path_temp)
                ## Return False to re-raise the exception
                return False
        ## If an exception occurs during the replacement, delete the temp path
        except Exception as e:
            ## Delete the temp path
            if Path(self.path_temp).exists() and self.delete_temp_on_error:
                delete_file_or_folder(self.path_temp)
            ## Raise the exception
            raise e
    
    def __str__(self):
        return f"SafeSaver(path_target={self.path_target}, path_temp={self.path_temp}, delete_temp_on_error={self.delete_temp_on_error})"
    

class MultiContextManager:
    """
    A context manager that allows multiple context managers to be used at once.
    RH 2024

    Args:
        *managers (context managers):
            Multiple context managers to be used at once.

    Demo:
        .. code-block:: python

            with MultiContextManager(
                torch.no_grad(), 
                temp_set_attr(obj, attr, new_val), 
                open('file.txt', 'w') as f,
            ):
                # do something
    """
    def __init__(self, *managers):
        self.managers = managers
        self.stack = ExitStack()

    def __enter__(self):
        for manager in self.managers:
            self.stack.enter_context(manager)

    def __exit__(self, exc_type, exc_value, traceback):
        self.stack.__exit__(exc_type, exc_value, traceback)
        

class SafeSaver(MultiContextManager):
    """
    Context manager for safely saving files or directories using a temporary
    path and path locking. This class combines the FileLock_AutoDeleting and
    AtomicSaver classes.
    RH 2024

    Args:
        path_target (Union[str, Path]):
            The path to the target file or directory to save.
        path_temp (Optional[Union[str, Path]):
            The path to the temporary file or directory to save. If None, it
            will be the target path with '.tmp' appended.
        path_lock (Optional[Union[str, Path]):
            The path to the lock file. If None, it will be the target path
            with '.lock' appended.
        overwrite (bool):
            If True, the target path will be overwritten if it exists.
        safe_save (bool):
            If True, the save operation will be wrapped in an AtomicSaver. If
            False and overwrite is True, the target path will be deleted before
            saving.
        delete_temp_on_error (bool):
            If True, the temporary path will be deleted if an error occurs.
        safe_save (bool):
            If True, the save operation will be wrapped in an AtomicSaver. If
            False and overwrite is True, the target path will be deleted before
            saving.
        file_lock (bool):
            If True, a lock file will be created to prevent concurrent access.
        timeout_lock (float):
            The timeout for acquiring the lock file.
        force_acquire_lock (bool):
            If True, the lock file will be deleted before acquiring the lock.
        force_release_lock (bool):
            If True, the lock file will be deleted before releasing the lock.

    Demo:
        .. code-block:: python

            with SafeSaver(
                path_target='file.txt', 
                path_temp='file.txt.tmp',
                overwrite=True, 
                safe_save=True, 
                delete_temp_on_error=False,
                file_lock=True,
                timeout_lock=5,
                force_acquire_lock=False,
                force_release_lock=False,
            ):
                save_object(path_temp, obj)
    """
    def __init__(
        self, 
        path_target: Union[str, Path], 
        path_temp: Optional[Union[str, Path]] = None,
        path_lock: Optional[Union[str, Path]] = None,
        overwrite: bool = False,
        delete_temp_on_error: bool = False,
        safe_save: bool = True,
        file_lock: bool = True,
        timeout_lock: float = 5,
        force_acquire_lock: bool = False,
        force_release_lock: bool = False,
    ):
        self.path_target = str(path_target)
        self.path_temp = str(path_temp) if path_temp is not None else self.path_target + '.tmp'
        self.path_lock = str(path_lock) if path_lock is not None else self.path_target + '.lock'

        self.overwrite = overwrite
        self.safe_save = safe_save
        self.delete_temp_on_error = delete_temp_on_error

        # Overwrite protection
        if not self.overwrite:
            if Path(self.path_target).exists():
                raise FileExistsError(f"Path already exists: {self.path_target} and overwrite is False.")

        ## Create managers list
        managers = []

        ## Create the lock file
        if file_lock:
            self.lock = FileLock_AutoDeleting(
                self.path_lock, 
                timeout=timeout_lock, 
                force=force_release_lock, 
                force_acquire=force_acquire_lock,
            )
            managers.append(self.lock)

        ## Create the atomic saver
        if self.safe_save:
            self.atomic_saver = AtomicSaver(
                path_target=self.path_target,
                path_temp=self.path_temp,
                overwrite=overwrite,
                delete_temp_on_error=delete_temp_on_error,
            )
            managers.append(self.atomic_saver)
        else:
            ## If not safe_save, delete the target path if it exists
            if self.overwrite:
                if Path(self.path_target).exists():
                    delete_file_or_folder(self.path_target)
            ## Set path_temp to path_target
            self.path_temp = self.path_target

        ## Initialize the MultiContextManager
        super().__init__(*managers)

    def __enter__(self):
        """
        Enters the context manager and returns the path to the temporary file.
        """
        super().__enter__()
        return self.path_temp

    def __str__(self):
        return f"SafeSaver(path_target={self.path_target}, path_temp={self.path_temp}, path_lock={self.path_lock}, overwrite={self.overwrite}, safe_save={self.safe_save}, delete_temp_on_error={self.delete_temp_on_error})"


# Import necessary modules
import tempfile
from pathlib import Path
import copy
import os
import math
import warnings

import pytest
import hypothesis
import richfile as rf


# Module-level variable for test data
DATA_BASIC = {
    "name": "John Doe",
    "age": 25,
    "address": {
        "street": "1234 Elm St",
        "zip": None,
    },
    "siblings": [
        "Jane",
        "Jim",
    ],
    "data": [1, 2, 3],
    (1, 2, 3): "complex key",
}

def save_and_load(file_path, obj):
    # Save data
    r = rf.RichFile(file_path).save(obj)
    # Load data back
    return rf.RichFile(file_path).load()


def cleanup(path):
    # Works with both files and directories
    if Path(path).is_file():
        Path(path).unlink(missing_ok=True)
    elif Path(path).is_dir():
        Path(path).rmdir(missing_ok=True)


def assert_equivalence(original, loaded):
    # Make closure for recursive comparison
    def compare(original, loaded):
        if isinstance(original, dict):
            assert isinstance(loaded, dict)
            assert set(original.keys()) == set(loaded.keys())
            for key in original:
                compare(original[key], loaded[key])
        elif isinstance(original, (list, tuple)):
            assert len(original) == len(loaded)
            for a, b in zip(original, loaded):
                compare(a, b)
        elif isinstance(original, (set, frozenset)):
            assert original == loaded
        elif isinstance(original, float) and math.isnan(original):
            assert isinstance(loaded, float) and math.isnan(loaded)
        else:
            assert original == loaded

    compare(original, loaded)


# Basic tests for RichFile functionality
def test_basic_save_load():
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_data.richfile"
        loaded_data = save_and_load(file_path, DATA_BASIC)
        assert_equivalence(DATA_BASIC, loaded_data)


def test_lazy_loading():
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_data_lazy.richfile"
        # Save data
        r = rf.RichFile(file_path).save(DATA_BASIC)
        # Lazy load a part of the data
        r = rf.RichFile(file_path)
        first_sibling = r["siblings"][0].load()
        assert first_sibling == "Jane"


# Property-based tests with hypothesis
# dict
@hypothesis.settings(deadline=600)
@hypothesis.given(
    hypothesis.strategies.dictionaries(
        keys=hypothesis.strategies.text()
        .filter(lambda x: len([c for c, v in rf.invalid_chars_filename(x).items() if not v]) == 0)
        .filter(lambda x: x not in rf.WINDOWS_RESERVED_NAMES),  ## Prevent Windows reserved names
        values=hypothesis.strategies.one_of(
            hypothesis.strategies.text(),
            hypothesis.strategies.integers(),
            hypothesis.strategies.floats(allow_nan=False),
            hypothesis.strategies.booleans(),
            hypothesis.strategies.none(),
        ),
    )
    .filter(lambda d: len({k.lower() for k in d.keys()}) == len(d)),  ## Prevent case-only sibling / duplicate keys
)
def test_save_load_dict(data):
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_data_dict.richfile"
        loaded_data = save_and_load(file_path, data)
        assert_equivalence(data, loaded_data)


# list
@hypothesis.settings(deadline=600)
@hypothesis.given(hypothesis.strategies.lists(hypothesis.strategies.integers()))
def test_save_load_list(data):
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_list_data.richfile"
        loaded_data = save_and_load(file_path, data)
        assert_equivalence(data, loaded_data)


# tuple
@hypothesis.settings(deadline=600)
@hypothesis.given(
    hypothesis.strategies.tuples(
        hypothesis.strategies.booleans(),
        hypothesis.strategies.floats(allow_nan=False),
        hypothesis.strategies.text(),
    )
)
def test_save_load_tuple(data):
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_tuple_data.richfile"
        loaded_data = save_and_load(file_path, data)
        assert_equivalence(data, loaded_data)


# set
@hypothesis.settings(deadline=600)
@hypothesis.given(hypothesis.strategies.sets(hypothesis.strategies.integers()))
def test_save_load_set(data):
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_set_data.richfile"
        loaded_data = save_and_load(file_path, data)
        assert_equivalence(data, loaded_data)


# frozenset
@hypothesis.settings(deadline=600)
@hypothesis.given(hypothesis.strategies.frozensets(hypothesis.strategies.integers()))
def test_save_load_frozenset(data):
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_frozenset_data.richfile"
        loaded_data = save_and_load(file_path, data)
        assert_equivalence(data, loaded_data)


# Test for registering new class types in various ways
def test_register_new_class_global():
    class CustomClass:
        def __init__(self, value):
            self.value = value

        def __eq__(self, other):
            if isinstance(other, CustomClass):
                return self.value == other.value
            return False

    # Register the new class type
    rf.functions.register_type(
        type_name="custom_class",
        function_load=lambda path: CustomClass(value=int(Path(path).read_text())),
        function_save=lambda path, obj: Path(path).write_text(str(obj.value)),
        object_class=CustomClass,
        library="python",
        suffix="txt",
    )

    # Check if type in typelookup
    type_lookup = rf.functions.TypeLookup()
    assert "custom_class" in type_lookup

    # Save and load data
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_custom_class.richfile"
        custom_data = {"data": CustomClass(42)}
        loaded_data = save_and_load(file_path, custom_data)
        assert loaded_data['data'] == custom_data['data']

    # Remove the type from typelookup
    rf.functions.remove_type("custom_class")
    type_lookup = rf.functions.TypeLookup()
    assert "custom_class" not in type_lookup


def test_register_sparse_coo_global():
    import sparse
    import numpy as np

    # Register sparse.COO type
    rf.functions.register_type(
        type_name="sparse_coo",
        function_load=lambda path: sparse.load_npz(path),
        function_save=lambda path, obj: sparse.save_npz(path, obj),
        object_class=sparse.COO,
        library="sparse",
        suffix="npz",
    )

    # Check if type in typelookup
    type_lookup = rf.functions.TypeLookup()
    assert "sparse_coo" in type_lookup

    # Save and load data
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_sparse_coo.richfile"
        custom_data = {"data": sparse.COO.from_numpy(np.random.rand(10, 10))}
        loaded_data = save_and_load(file_path, custom_data)
        assert np.allclose(loaded_data['data'].todense(), custom_data['data'].todense())

    # Remove the type from typelookup
    rf.functions.remove_type("sparse_coo")
    type_lookup = rf.functions.TypeLookup()
    assert "sparse_coo" not in type_lookup


def test_register_nparray_local():
    import numpy as np

    # Prepare temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        path_rf = Path(temp_dir) / "test_nparray_local.richfile"
        data = {"data": np.random.rand(10, 10)}
        # Make a new RichFile object
        r = rf.RichFile(path_rf)
        # Register the type locally
        ## Make a new TypeLookup dict
        tl = {
            "type_name": "nparray",
            "function_load": lambda path: np.load(path),
            "function_save": lambda path, obj: np.save(path, obj),
            "object_class": np.ndarray,
            "library": "numpy",
            "suffix": "npy",
            "versions_supported": [],
        }
        ## First make sure it is not in the global typelookup
        assert "nparray" not in rf.functions.TypeLookup()
        ## Register the type locally
        r.register_type(**tl)
        # Check if type in typelookup
        assert "nparray" in r.type_lookup
        # Remove the type from typelookup
        r.remove_type("nparray")
        # Check if type in typelookup
        assert "nparray" not in r.type_lookup
        # Add it back using a dictionary
        r.register_type_from_dict(tl)
        # Check if type in typelookup
        assert "nparray" in r.type_lookup

        # Save and load data
        r.save(data)
        loaded_data = r.load()

        # Check if the data is the same
        assert np.allclose(data['data'], loaded_data['data'])

        # Make sure global typelookup is not affected
        assert "nparray" not in rf.functions.TypeLookup()


# Error handling tests
def test_register_new_class_missing_argument():
    class CustomClass:
        pass

    with pytest.raises(TypeError):
        rf.functions.register_type(
            type_name="custom_class",
            function_load=lambda path: CustomClass(),  # Missing function_save
            object_class=CustomClass,
            suffix="txt",
            library="python",
            versions_supported=[],
        )
def test_register_new_class_invalid_type():
    class CustomClass:
        pass

    with pytest.raises(TypeError):
        rf.functions.register_type(
            type_name="custom_class",
            function_load="not a function",  # Invalid type for function_load
            function_save=lambda path, obj: None,
            object_class=CustomClass,
            suffix="txt",
            library="python",
            versions_supported=[],
        )
def test_register_new_class_type_collision():
    class CustomClass:
        pass

    rf.functions.register_type(
        type_name="custom_class",
        function_load=lambda path: CustomClass(),
        function_save=lambda path, obj: None,
        object_class=CustomClass,
        suffix="txt",
        library="python",
        versions_supported=[],
    )

    with pytest.raises(KeyError):
        rf.functions.register_type(
            type_name="custom_class",
            function_load=lambda path: int,  # Same name, different type
            function_save=lambda path, obj: None,
            object_class=int,
            suffix="txt",
            library="python",
            versions_supported=[],
        )

    rf.functions.remove_type("custom_class")
def test_remove_nonexistent_type():
    with pytest.raises(KeyError):
        rf.functions.remove_type("nonexistent_type")


def test_save_unsupported_type():
    with pytest.raises(TypeError):
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = Path(temp_dir) / "test_unsupported_type.richfile"
            rf.RichFile(file_path).save(math)

def test_load_corrupted_file():
    import json
    with tempfile.TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "corrupted_file.richfile"
        ## Save good json data
        r = rf.RichFile(path).save({"data": 42})
        ## Corrupt the file
        with open(r['data'].path, "w") as f:
            ## Overwrite the file with bad data
            f.write("bad data")
        ## Load the file
        with pytest.raises(json.JSONDecodeError):
            r.load()

# Edge case tests
def test_save_empty_dict():
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_empty_dict.richfile"
        rf.RichFile(file_path).save({})
        loaded_data = rf.RichFile(file_path).load()
        assert loaded_data == {}

def test_save_nested_data_structures():
    nested_data = {
        "deep_lists": [[[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12], [[[[[[13,]]]]]]]]],
        "deep_dicts": {"a": {"b": {"c": {"d": {"e": {"f": 42}}}}},},
    }
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_nested_data.richfile"
        rf.RichFile(file_path).save(nested_data)
        loaded_data = rf.RichFile(file_path).load()
        assert loaded_data == nested_data

def test_register_type_overwrite():
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_local_type_overwrite.richfile"
        td = {
            "type_name": "custom_type",
            "object_class": type(None),
            "function_load": lambda path: None,
            "function_save": lambda path, obj: None,
            "suffix": "txt",
            "library": "python",
            "versions_supported": [],
        }
        r = rf.RichFile(file_path)
        r.register_type(**td)
        with pytest.raises(KeyError):
            r.register_type(**td)
        ## Now go the other way
        rf.functions.register_type(**td)
        with pytest.raises(KeyError):
            r = rf.RichFile(file_path)
            r.register_type(**td)
        ## cleanup
        rf.functions.remove_type("custom_type")
        
# Windows reserved names should raise a clear ValueError
def test_save_load_windows_reserved_name(tmp_path):
    file_path = tmp_path / "test_windows_reserved.richfile"
    for name in rf.WINDOWS_RESERVED_NAMES:
        with pytest.raises(ValueError) as excinfo:
            rf.RichFile(
                path=file_path, 
                overwrite=True, 
                check=True, 
                name_dict_items=True
            ).save({name: "value"})
        assert f"Cannot save folder name" in str(excinfo.value), f"Expected ValueError for {name} not raised. Found: {excinfo.value}"


# Case-only sibling keys should raise a ValueError on case-insensitive filesystems
def test_save_load_case_only_siblings(tmp_path):
    file_path = tmp_path / "test_case_siblings.richfile"
    data = {"foo": "value1", "FOO": "value2"}
    # Expect a ValueError about case-only collisions
    with pytest.raises(ValueError) as excinfo:
        rf.RichFile(
            path=file_path,
            overwrite=True,
            check=True,
            name_dict_items=True
        ).save(data)
    assert "Found case-only" in str(excinfo.value)



__version__ = "0.5.1"

VERSIONS_RICHFILE_SUPPORTED = [">=0.3.1", "<1.0.0"]
PYTHON_VERSIONS_SUPPORTED = [">=3", "<4"]

FILENAME_METADATA   = ".metadata.richfile"
FILENAME_TYPELOOKUP = ".typelookup.richfile"

JSON_INDENT = 4

WINDOWS_RESERVED_NAMES = {
    "CON", "PRN", "AUX", "NUL",
    *{f"COM{i}" for i in range(1, 10)},
    *{f"LPT{i}" for i in range(1, 10)},
}


## Import important stuff from util.py into top-level namespace
from . import functions, util
from .util import RichFile, load_folder, load_element, load_folder_metadata, save_object, invalid_chars_filename
from . import demo


from typing import Union, List, Dict, Any, Callable, Optional, Tuple
from pathlib import Path
import inspect
import warnings
import importlib

from . import util


_TYPE_LOOKUP = [
    {
        "type_name":          "list",
        "function_load":      util.load_folder,
        "function_save":      util.save_container,
        "object_class":       list,
        "suffix":             "list",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },
    {
        "type_name":          "tuple",
        "function_load":      util.load_folder,
        "function_save":      util.save_container,
        "object_class":       tuple,
        "suffix":             "tuple",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },
    {
        "type_name":          "set",
        "function_load":      util.load_folder,
        "function_save":      util.save_container,
        "object_class":       set,
        "suffix":             "set",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },
    {
        "type_name":          "frozenset",
        "function_load":      util.load_folder,
        "function_save":      util.save_container,
        "object_class":       frozenset,
        "suffix":             "frozenset",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },    
    {
        "type_name":          "dict",
        "function_load":      util.load_folder,
        "function_save":      util.save_container,
        "object_class":       dict,
        "suffix":             "dict",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },

    {
        "type_name":          "dict_item",
        "function_load":      util.load_folder,
        "function_save":      util.save_container,
        "object_class":       util.DictItem,
        "suffix":             "dict_item",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },

    {
        "type_name":          "float",
        "function_load":      util.load_float,
        "function_save":      util.save_json,
        "object_class":       float,
        "suffix":             "json",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },
    {
        "type_name":          "int",
        "function_load":      util.load_int,
        "function_save":      util.save_json,
        "object_class":       int,
        "suffix":             "json",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },
    {
        "type_name":          "str",
        "function_load":      util.load_str,
        "function_save":      util.save_json,
        "object_class":       str,
        "suffix":             "json",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },
    {
        "type_name":          "bool",
        "function_load":      util.load_bool,
        "function_save":      util.save_json,
        "object_class":       bool,
        "suffix":             "json",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },
    {
        "type_name":          "None",
        "function_load":      util.load_None,
        "function_save":      util.save_json,
        "object_class":         type(None),
        "suffix":             "json",
        "library":            "python",
        "versions_supported": [">=3", "<4"],
    },
]

_TYPE_KEYS_REQUIRED = [
    "type_name",
    "function_load",
    "function_save",
    "object_class",
    "suffix",
    "library",
    "versions_supported",
]

class TypeLookup:
    """
    A class for looking up the properties of the different types of objects that
    can be saved and loaded.

    This class behaves sort of like a dictionary, but instead of inputting a
    single key and getting a single output, you input a tuple corresponding to a
    ``(key, value)`` pair and return a list of property dictionaries that
    contain that matching key-value pair.

    The class also has methods for adding and removing properties.

    Args:
        type_lookup (List[Dict], optional): 
            A list of dictionaries containing the properties of the objects.
            Defaults to functions._TYPE_LOOKUP.
    """
    def __init__(self, type_lookup: List[Dict] = _TYPE_LOOKUP):
        self.properties = type_lookup
        _verify_validity_of_type_lookup(self.properties)

    def add_property(self, prop: Dict) -> None:
        """
        Adds a property to the list of properties.
        """
        self.properties.append(prop)
        _verify_validity_of_type_lookup(self.properties)

    def remove_property(self, type_name: str) -> None:
        """
        Removes a property from the list of properties.
        """
        self.properties = [prop for prop in self.properties if prop["type_name"] != type_name]

    def __len__(self) -> int:
        return len(self.properties)
    
    def __iter__(self):
        return iter(self.properties)
    
    def __repr__(self) -> str:
        return f"TypeLookup({[str(prop['type_name']) for prop in self.properties]})"
    
    def __add__(self, prop: Dict) -> None:
        """
        Adds a property to the list of properties.

        Args:
            prop (Dict): 
                A dictionary containing the properties of the object.
        """
        if not isinstance(prop, dict):
            raise ValueError("prop must be a dictionary.")
        self.add_property(prop)
        _verify_validity_of_type_lookup(self.properties)

    def __sub__(self, type_name: str) -> None:
        """
        Removes a property from the list of properties.

        Args:
            type_name (str): 
                The name of the type to remove.
        """
        if not isinstance(type_name, str):
            raise ValueError("type_name must be a string.")
        self.remove_property(type_name)
        _verify_validity_of_type_lookup(self.properties)

    def __contains__(self, query: Union[str, type]) -> bool:
        if isinstance(query, str):
            return query in [prop["type_name"] for prop in self.properties]
        elif isinstance(query, type):
            return query in [prop["object_class"] for prop in self.properties]

    def __getitem__(self, query: Union[str, type, Tuple[str]]) -> Union[Dict, List[Dict]]:
        """
        Returns the property dictionary that contain the matching key-value
        pair. If multiple properties match the key-value pair, a warning is
        raised and a list of property dictionaries is returned.

        Args:
            key_value (Union[str, type, Tuple[str]]): 
                The key-value pair to match. Can be one of the following input types: \n
                    * ``str``: Return prop dict where ``prop["type_name"] == key_value``
                    * ``type``: Return prop dict where ``prop["object_class"] ==
                      key_value``
                    * ``Tuple[str]``: Return prop dict where ``prop[key_value[0]] ==
                      key_value[1]``

        Returns:
            Union[Dict, List[Dict]]:
                The property dictionary or list of property dictionaries that
                match the key-value pair input.
        """
        if isinstance(query, tuple):
            if len(query) != 2:
                raise ValueError("key_value must be a tuple of length 2.")
            if not isinstance(query[0], str):
                raise ValueError("The first element of key_value must be a string.")
            return [prop for prop in self.properties if prop[query[0]] == query[1]]
        
        elif isinstance(query, str):
            for prop in self.properties:
                if prop["type_name"] == query:
                    return prop
            raise ValueError(f"String '{query}' not found in 'type_name' fields of available properties. Available properties: {[prop['type_name'] for prop in self.properties]}")
        
        elif isinstance(query, type):
            ## Go through the type_lookup dictionary to find the type of the object. isinstance checks for inherited classes as well.
            ### Check with 'is' first to avoid inheritance issues (ex. bool is a subclass of int)
            ### Then check if available "object_class" types are superclasses of the query type
            ### Then check using string comparison
            for props in self.properties:   
                if query is props["object_class"]:
                    return props
            for props in self.properties:
                if issubclass(query, props["object_class"]):
                    return props
            for props in self.properties:
                if str(query) == str(props["object_class"]):
                    return props
            ### Else, raise an error
            raise TypeError(f"Type '{query}' not supported. You can register a new type using `richfile.functions.register_type`.")

        else:
            raise ValueError("key_value must be a string, type, or tuple of length 2.")
        
    def __setitem__(self, key: str, props: Dict) -> None:
        """
        Update the property dictionary that has the field ``"type_name"`` that
        matches input arg ``key`` with items from the dictionary ``props``.

        Args:
            key (str): 
                The ``"type_name"`` field to match.
            props (Dict): 
                The new items to update the property dictionary with.
        """
        if not isinstance(key, str):
            raise ValueError("key must be a string.")
        if not isinstance(props, dict):
            raise ValueError("props must be a dictionary.")
        ## Make sure keys in props are a subset of keys in _TYPE_KEYS_REQUIRED
        if not set(props.keys()).issubset(set(_TYPE_KEYS_REQUIRED)):
            keys_extra = set(props.keys()) - set(_TYPE_KEYS_REQUIRED)
            raise ValueError(f"Extra keys in props: {keys_extra}. Required keys: {_TYPE_KEYS_REQUIRED}")
        
        ## Update the property dictionary
        for prop in self.properties:
            if prop["type_name"] == key:
                prop.update(props)
                _verify_validity_of_type_lookup(self.properties)
                return
    
    def __delitem__(self, key: str) -> None:
        """
        Removes the property dictionary that has the field ``"type_name"`` that
        matches input arg ``key``.

        Args:
            key (str): 
                The ``"type_name"`` field to match.
        """
        if not isinstance(key, str):
            raise ValueError("key must be a string.")
        self.remove_property(key)
        _verify_validity_of_type_lookup(self.properties)

    def types_metadata(self) -> List[str]:
        """
        Returns information about each type in the type lookup.
        Functions and classes are converted to strings.
        """
        import inspect
        types_metadata = []
        for prop in self.properties:
            prop_metadata = {}
            for key, value in prop.items():
                if isinstance(value, type):
                    prop_metadata[key] = str(value)
                elif callable(value):
                    prop_metadata[key] = inspect.getsource(value)
                elif isinstance(value, list):
                    if all([isinstance(elem, str) for elem in value]):
                        prop_metadata[key] = value
                    else:
                        raise ValueError(f"Unexpected item in self.properties: {value}")
                else:
                    if isinstance(value, str):
                        prop_metadata[key] = value
                    else:
                        raise ValueError(f"Unexpected item in self.properties: {value}")
            types_metadata.append(prop_metadata)

        return types_metadata


def _verify_validity_of_type_lookup(type_lookup: List[Dict]) -> None:
    """
    Verifies that the properties are valid.
    
    Rules:
    - Each property dict must have exactly the following keys and
    corresponding types:
        - type_name: str
        - function_load: Callable
        - function_save: Callable
        - object_class: type
        - suffix: str
        - library: str
        - versions_supported: List[str]
    - The values of the 'type_name' and 'type_class' fields must be unique
        across all properties.
    - The 'versions_supported' field may be empty, but if it is not, it must
        be formatted like: [">=0", "<3.1"] and work with util.is_version_compatible.
    """
    ## Check that each property dict has the correct keys. Use sets
    for prop in type_lookup:
        for prop in type_lookup:
            if set(prop.keys()) != set(_TYPE_KEYS_REQUIRED):
                keys_missing = set(_TYPE_KEYS_REQUIRED) - set(prop.keys())
                keys_extra = set(prop.keys()) - set(_TYPE_KEYS_REQUIRED)
                raise ValueError(f"Property dict doesn't have matching keys to required keys. Missing keys: {keys_missing}. Extra keys: {keys_extra}. Required keys: {_TYPE_KEYS_REQUIRED}")
    
    ## Check that the 'type_name' and 'type_class' fields are unique. Use sets
    type_names = [prop["type_name"] for prop in type_lookup]
    if len(type_names) != len(set(type_names)):
        type_names_duplicates = {name: type_names.count(name) for name in type_names if type_names.count(name) > 1}
        raise ValueError(f"Duplicate type names and their counts: {type_names_duplicates}. Call `richfile.functions.remove_type` to remove duplicates.")
    
    ## Check that the 'versions_supported' field is formatted correctly
    for prop in type_lookup:
        ## Check formatting first
        if not isinstance(prop["versions_supported"], list):
            raise ValueError(f"versions_supported must be a list of strings.")
        if not all([isinstance(version, str) for version in prop["versions_supported"]]):
            raise ValueError(f"versions_supported must be a list of strings.")

def _verify_validity_of_new_type(prop: Dict) -> Dict:
    """
    Verifies that the properties of a new type are valid.

    Args:
        prop (Dict): 
            The properties of the new type.
            Must contain the following keys and corresponding types:
                - type_name: str
                - function_load: Callable
                - function_save: Callable
                - object_class: type
                - suffix: str
                - library: str
                - versions_supported: List[str]

    Returns:
        Dict: 
            Same dict as input, but with any necessary corrections.
    """
    ## Check that prop is a dictionary
    if not isinstance(prop, dict):
        raise ValueError("prop must be a dictionary.")
    
    ## Set versions_supported to empty list if not provided
    if "versions_supported" not in prop.keys():
        prop["versions_supported"] = []

    ## Check that the property dict has the correct keys. Use sets
    if set(prop.keys()) != set(_TYPE_KEYS_REQUIRED):
        keys_missing = set(_TYPE_KEYS_REQUIRED) - set(prop.keys())
        keys_extra = set(prop.keys()) - set(_TYPE_KEYS_REQUIRED)
        raise ValueError(f"Property dict doesn't have matching keys to required keys. Missing keys: {keys_missing}. Extra keys: {keys_extra}. Required keys: {_TYPE_KEYS_REQUIRED}")
    
    type_name, \
    function_load, \
    function_save, \
    object_class, \
    suffix, \
    library, \
    versions_supported = [prop[key] for key in [
        "type_name",
        "function_load",
        "function_save",
        "object_class",
        "suffix",
        "library",
        "versions_supported",
    ]]   

    ## Clean the inputs
    ### type_name
    if not isinstance(type_name, str):
        raise TypeError("`type_name` must be a string.")
    #### empty
    if type_name == "":
        raise ValueError("`type_name` cannot be empty.")
    #### check filename safety
    util._check_filename_safety(name=type_name, warn=True, raise_error=False)
    #### check for duplicates
    if type_name in [prop["type_name"] for prop in _TYPE_LOOKUP]:
        raise KeyError(f"Type {type_name} already registered. Consider using `richfile.functions.remove_type` to remove it.")
    ### function_load
    if not callable(function_load):
        raise TypeError("`function_load` must be a callable.")
    util.functions._check_function_args(func=function_load, args=["path"])
    ### function_save
    if not callable(function_save):
        raise TypeError("`function_save` must be a callable.")
    util.functions._check_function_args(func=function_save, args=["obj", "path",])
    ### object_type
    if not isinstance(object_class, type):
        raise TypeError("`object_type` must be a type.")
    ### suffix
    if not isinstance(suffix, str):
        raise TypeError("`suffix` must be a string.")
    util._check_filename_safety(name=suffix, warn=True, raise_error=True)
    ### library
    if not isinstance(library, str):
        raise TypeError("`library` must be a string.")
    #### empty
    if library == "":
        raise ValueError("`library` cannot be empty.")
    #### make sure it's a valid library
    if library not in ["python", "builtins"]:
        try:
            importlib.import_module(library)
        except ImportError:
            warnings.warn(f"Library {library} not found or not importable in the current environment.")
    ### versions_supported
    if not isinstance(versions_supported, list):
        raise TypeError("`versions_supported` must be a list.")
    #### check that all elements are strings
    if not all(isinstance(version, str) for version in versions_supported):
        raise TypeError("All elements in `versions_supported` must be strings.")
    #### check that environment version is supported
    if len(versions_supported) > 0:
        version = util._get_library_version(library=library)
        util.is_version_compatible(version=version, rules=versions_supported)

    ## remove leading '.' and check for invalid characters
    if suffix[0] == ".":
        suffix = suffix[1:]
    
    ## return the cleaned properties
    return {
        "type_name":          type_name,
        "function_load":      function_load,
        "function_save":      function_save,
        "object_class":       object_class,
        "suffix":             suffix,
        "library":            library,
        "versions_supported": versions_supported,
    }


def register_type(
    type_name: str,
    function_load: Callable,
    function_save: Callable,
    object_class: type,
    suffix: str,
    library: str,
    versions_supported: List[str] | None = []
):
    """
    Registers a new type of object that can be saved and loaded.

    Args:
        type_name (str): 
            The name of the type.
        function_load (Callable): 
            The function that loads the object from a file.
        function_save (Callable): 
            The function that saves the object to a file.
        object_class (type): 
            The class of the object.
        suffix (str): 
            The suffix of the file.
        library (str): 
            The library that the object is associated with.
        versions_supported (List[str], optional): 
            A list of version strings that are supported by the object. 
            If empty, all versions are supported. Defaults to [].
    """
    prop = {
        "type_name":          type_name,
        "function_load":      function_load,
        "function_save":      function_save,
        "object_class":       object_class,
        "suffix":             suffix,
        "library":            library,
        "versions_supported": versions_supported,
    }
    register_type_from_dict(prop)

def register_type_from_dict(prop: Dict) -> None:
    """
    Registers a new type of object that can be saved and loaded.

    Args:
        prop (Dict): 
            A dictionary containing the properties of the object.
            Must contain the following keys and corresponding types:
                - type_name: str
                - function_load: Callable
                - function_save: Callable
                - object_class: type
                - suffix: str
                - library: str
                - versions_supported: List[str] (optional)
    """
    ## Clean the inputs
    prop = _verify_validity_of_new_type(prop)
    _TYPE_LOOKUP.append(prop)
    _verify_validity_of_type_lookup(_TYPE_LOOKUP)

def remove_type(type_name: str) -> None:
    ## Make sure the type_name is valid
    if type_name not in [prop["type_name"] for prop in _TYPE_LOOKUP]:
        raise KeyError(f"Type {type_name} not found.")
    ## Find the index of the type_name
    idx_to_remove = [idx for idx, prop in enumerate(_TYPE_LOOKUP) if prop["type_name"] == type_name][0]
    ## Remove the type_name
    _TYPE_LOOKUP.pop(idx_to_remove)


class Container:
    def __init__(
        self,
        type_name: str,
        object_class: type,
        suffix: str,
        library: str,
        versions_supported: List[str] | None = [],
    ):
        self.type_name = type_name
        self.object_class = object_class
        self.suffix = suffix
        self.library = library
        self.versions_supported = versions_supported

    def register_type(self):
        register_type(
            type_name=self.type_name,
            function_load=self.function_load,
            function_save=self.function_save,
            object_class=self.object_class,
            suffix=self.suffix,
            library=self.library,
            versions_supported=self.versions_supported,
        )

    def function_save(
        self,
        obj: Any,
        path: Union[str, Path],
        type_lookup: Dict,
        check: bool = True,
        overwrite: bool = False,
        name_dict_items: bool = True,
        **kwargs,
    ) -> None:
        """
        Saves an object to the given path as a container.
        """
        if not isinstance(obj, self.object_class):
            raise TypeError(f"Object must be of type {self.object_class}.")
        
        util.save_container(
            obj=obj.__dict__,
            path=path,
            type_name=self.type_name,
            type_lookup=type_lookup,
            check=check,
            overwrite=overwrite,
            name_dict_items=name_dict_items,
        )

    def function_load(
        self,
        path: Union[str, Path],
        type_lookup: Dict,
        check: bool = True,
        **kwargs,
    ) -> Any:
        """
        Loads an object from the given path.
        """
        out = self.object_class.__new__(self.object_class)
        out.__dict__ = util.load_folder(
            path=path,
            type_lookup=type_lookup,
            check=check,
            **kwargs,
        )
        return out
    
    def get_property_dict(self) -> Dict:
        return {
            "type_name":          self.type_name,
            "function_load":      self.function_load,
            "function_save":      self.function_save,
            "object_class":       self.object_class,
            "suffix":             self.suffix,
            "library":            self.library,
            "versions_supported": self.versions_supported,
        }
    

################################################################################################
######################################### CUSTOM TYPES #########################################
################################################################################################


#################
## numpy_array ##
#################

# import numpy as np

# def save_npy_array(
#     obj: np.ndarray,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves a NumPy array to the given path.
#     """
#     np.save(path, obj, **kwargs)

# def load_npy_array(
#     path: Union[str, Path],
#     **kwargs,
# ) -> np.ndarray:
#     """
#     Loads an array from the given path.
#     """    
#     return np.load(path, **kwargs)

# _TYPE_LOOKUP.append({
#         "type_name":          "numpy_array",
#         "function_load":      load_npy_array,
#         "function_save":      save_npy_array,
#         "object_class":       np.ndarray,
#         "suffix":             "npy",
#         "library":            "numpy",
#         "versions_supported": [">=1", "<3"],
#     })

##################
## numpy_scalar ##
##################

# import numpy as np

# _TYPE_LOOKUP.append({
#         "type_name":          "numpy_scalar",
#         "function_load":      load_npy_array,
#         "function_save":      save_npy_array,
#         "object_class":       np.number,
#         "suffix":             "npy",
#         "library":            "numpy",
#         "versions_supported": [">=1", "<3"],
#     })

##############
## datetime ##
##############

# import datetime

# def save_datetime(
#     obj: datetime.datetime,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves a datetime object to the given path.
#     """
#     util.save_json(obj.isoformat(), path, **kwargs)

# def load_datetime(
#     path: Union[str, Path],
#     **kwargs,
# ) -> datetime.datetime:
#     """
#     Loads a datetime object from the given path.
#     """
#     return datetime.datetime.fromisoformat(util.load_json(path, **kwargs))

# _TYPE_LOOKUP.append({
#         "type_name":          "datetime",
#         "function_load":      load_datetime,
#         "function_save":      save_datetime,
#         "object_class":       datetime.datetime,
#         "suffix":             "json",
#         "library":            "python",
#         "versions_supported": [">=3", "<4"],
#     })

########################
## scipy_sparse_array ##
########################

# import scipy.sparse

# def save_sparse_array(
#     obj: scipy.sparse.spmatrix,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves a SciPy sparse matrix to the given path.
#     """
#     scipy.sparse.save_npz(path, obj, **kwargs)

# def load_sparse_array(
#     path: Union[str, Path],
#     **kwargs,
# ) -> scipy.sparse.csr_matrix:
#     """
#     Loads a sparse array from the given path.
#     """        
#     return scipy.sparse.load_npz(path, **kwargs)

# _TYPE_LOOKUP.append({
#         "type_name":          "scipy_sparse_array",
#         "function_load":      load_sparse_array,
#         "function_save":      save_sparse_array,
#         "object_class":       scipy.sparse.spmatrix,
#         "suffix":             "npz",
#         "library":            "scipy",
#         "versions_supported": [">=1", "<2"],
#     })

##################
## torch_tensor ##
##################

# import torch
# import numpy as np

# def save_torch_tensor(
#     obj: torch.Tensor,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves a PyTorch tensor as a NumPy array to the given path.
#     """
#     np.save(path, obj.detach().cpu().numpy(), **kwargs)

# def load_torch_tensor(
#     path: Union[str, Path],
#     **kwargs,
# ) -> np.ndarray:
#     """
#     Loads a PyTorch tensor from the given path.
#     """
#     return torch.as_tensor(np.load(path, **kwargs))

# _TYPE_LOOKUP.append({
#         "type_name":          "torch_tensor",
#         "function_load":      load_torch_tensor,
#         "function_save":      save_torch_tensor,
#         "object_class":       torch.Tensor,
#         "suffix":             "npy",
#         "library":            "numpy",
#         "versions_supported": [">=1", "<3"],
#     })

##################
## optuna_study ##
##################

# import optuna

# def save_optuna_study(
#     obj: optuna.study.Study,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves an Optuna study to the given path.
#     """
#     import pickle
#     with open(path, "wb") as f:
#         pickle.dump(obj, f, **kwargs)

# def load_optuna_study(
#     path: Union[str, Path],
#     **kwargs,
# ) -> optuna.study.Study:
#     """
#     Loads an Optuna study from the given path.
#     """
#     import pickle
#     with open(path, "rb") as f:
#         return pickle.load(f, **kwargs)
    
# _TYPE_LOOKUP.append({
#         "type_name":          "optuna_study",
#         "function_load":      load_optuna_study,
#         "function_save":      save_optuna_study,
#         "object_class":       optuna.study.Study,
#         "suffix":             "optuna_study_pkl",
#         "library":            "python",
#         "versions_supported": [">=2", "<3"],
#     })

################
## pandas_csv ##
################

# import pandas as pd

# def save_pandas_csv(
#     obj: pd.DataFrame,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves a pandas DataFrame to the given path as a CSV.
#     """
#     obj.to_csv(path, **kwargs)


# def load_pandas_csv(
#     path: Union[str, Path],
#     **kwargs,
# ) -> pd.DataFrame:
#     """
#     Loads a pandas DataFrame from the given path.
#     """
#     return pd.read_csv(path, **kwargs)


# _TYPE_LOOKUP.append({
#         "type_name":          "pandas_csv",
#         "function_load":      load_pandas_csv,
#         "function_save":      save_pandas_csv,
#         "object_class":       pd.DataFrame,
#         "suffix":             "csv",
#         "library":            "pandas",
#         "versions_supported": [">=1", "<3"],
#     })

###################
## PIL_image_png ##
###################

# from PIL import Image

# def save_PIL_image_png(
#     obj: Image.Image,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves a PIL image to the given path as a PNG.
#     """
#     with open(path, "wb") as f:
#         obj.save(f, format="PNG", **kwargs)

# def load_PIL_image_png(
#     path: Union[str, Path],
#     **kwargs,
# ) -> Image.Image:
#     """
#     Loads a PIL image from the given path.
#     """
#     return Image.open(path, **kwargs)

# _TYPE_LOOKUP.append({
#         "type_name":          "PIL_image_png",
#         "function_load":      load_PIL_image_png,
#         "function_save":      save_PIL_image_png,
#         "object_class":       Image.Image,
#         "suffix":             "png",
#         "library":            "PIL",
#         "versions_supported": [">=6", "<11"],
#     })

#########################
## sklearn_model_skops ##
#########################

# import sklearn
# import skops.io

# def save_sklearn_model_skops(
#     obj: sklearn.base.BaseEstimator,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves a scikit-learn model to the given path using skops.
#     """
#     with open(path, "wb") as f:
#         skops.io.dump(obj, f, **kwargs)

# def load_sklearn_model_skops(
#     path: Union[str, Path],
#     **kwargs,
# ) -> sklearn.base.BaseEstimator:
#     """
#     Loads a scikit-learn model from the given path using skops.
#     """
#     with open(path, "rb") as f:
#         return skops.io.load(f, **kwargs)
    
# _TYPE_LOOKUP.append({
#         "type_name":          "sklearn_model_skops",
#         "function_load":      load_sklearn_model_skops,
#         "function_save":      save_sklearn_model_skops,
#         "object_class":       sklearn.base.BaseEstimator,
#         "suffix":             "skops",
#         "library":            "skops",
#         "versions_supported": [">=0", "<1"],
#     })

####################
## numpy_array_h5 ##
####################

# import tables

# def save_npy_array_h5(
#     obj: np.ndarray,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves a NumPy array to the given path using HDF5.
#     """
#     with tables.open_file(path, "w") as f:
#         f.create_array("/", "data", obj)

# def load_npy_array_h5(
#     path: Union[str, Path],
#     **kwargs,
# ) -> np.ndarray:
#     """
#     Loads an array from the given path using HDF5.
#     """
#     with tables.open_file(path, "r") as f:
#         return f.root.data.read()

# _TYPE_LOOKUP.append({
#         "type_name":          "numpy_array_h5",
#         "function_load":      load_npy_array_h5,
#         "function_save":      save_npy_array_h5,
#         "object_class":       np.ndarray,
#         "suffix":             "h5",
#         "library":            "tables",
#         "versions_supported": [">=3", "<4"],
#     })

################
## pickle_obj ##
################

# import pickle

# def save_pickle_obj(
#     obj: Any,
#     path: Union[str, Path],
#     **kwargs,
# ) -> None:
#     """
#     Saves an object to the given path using pickle.
#     """
#     with open(path, "wb") as f:
#         pickle.dump(obj, f, **kwargs)

# def load_pickle_obj(
#     path: Union[str, Path],
#     **kwargs,
# ) -> Any:
#     """
#     Loads an object from the given path using pickle.
#     """
#     with open(path, "rb") as f:
#         return pickle.load(f, **kwargs)
    
# _TYPE_LOOKUP.append({
#         "type_name":          "pickle_obj",
#         "function_load":      load_pickle_obj,
#         "function_save":      save_pickle_obj,
#         "object_class":       object,
#         "suffix":             "pkl",
#         "library":            "pickle",
#         "versions_supported": [],
#     })

###############################################################################################

def _check_function_args(func: Callable, args: List[str]):
    """
    Checks functions called by load_element to make sure they accept the correct
    arguments.

    Args:
        func (Callable): 
            The function to check.
        args (List[str]): 
            The arguments that the function should accept
    """
    sig = inspect.signature(func)
    if not all([param in sig.parameters for param in args]):
        raise ValueError(f"Function {func.__name__} does not accept the correct arguments. Expected: {args}. Got: {list(sig.parameters.keys())}. Often, add ``**kwargs`` to the function definition.")

[_check_function_args(func=props["function_load"], args=["path",])        for props in _TYPE_LOOKUP]
[_check_function_args(func=props["function_save"], args=["obj", "path",]) for props in _TYPE_LOOKUP]